from autolens.data import ccd
from autolens.data.array import mask as msk
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy as g
from autolens.lens import ray_tracing
from autolens.lens import lens_fit
from autolens.lens import lens_data as ld
from autolens.lens import plane as pl
from autolens.model.inversion import inversions as inv
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.lens.plotters import lens_fit_plotters

# In this tutorial, we'll introduce a new pixelization, called an adaptive-pixelization. This pixelization doesn't use
# uniform grid of rectangular pixels, but instead uses irregular 'Voronoi' pixels. So, why would we want to do that?
# Lets take another look at the rectangular grid, and think about its weakness.

# This is the image we're used to fitting with inversions now.
def simulate():

    from autolens.data.array import grids
    from autolens.model.galaxy import galaxy as g
    from autolens.lens import ray_tracing

    psf = ccd.PSF.simulate_as_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    image_plane_grid_stack = grids.GridStack.grid_stack_for_simulation(shape=(150, 150), pixel_scale=0.05,
                                                                       psf_shape=(11, 11))

    lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0,
                                                        einstein_radius=1.6))
    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.7, phi=135.0, intensity=0.2,
                                                       effective_radius=0.2, sersic_index=2.5))
    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grid_stack=image_plane_grid_stack)

    return ccd.CCDData.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.05,
                                        exposure_time=300.0, psf=psf, background_sky_level=0.1, add_noise=True)

# Lets quickly remind ourselves of the image, and the 3.0" circular mask we'll use to mask it.
ccd_data = simulate()
mask = msk.Mask.circular(shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.0)
# ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data, mask=mask)

# The lines of code below do everything we're used to, that is, setup an image and its grid stack, mask it, trace it
# via a tracer, setup the rectangular mapper, etc.
lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0,
                                                    einstein_radius=1.6))

lens_data = ld.LensData(ccd_data=ccd_data, mask=mask)

adaptive = pix.AdaptiveMagnification(shape=(10, 20))
image_plane_pix_grid = adaptive.image_plane_pix_grid_from_regular_grid(regular_grid=lens_data.grid_stack.regular)

image_plane_grid_stack = lens_data.grid_stack.new_grid_stack_with_pix_grid_added(pix_grid=image_plane_pix_grid.sparse_grid,
                                                                                 regular_to_nearest_pix=image_plane_pix_grid.regular_to_sparse)

image_plane = pl.Plane(grid_stack=image_plane_grid_stack, galaxies=[lens_galaxy])
source_plane_grid_stack = image_plane.trace_grid_stack_to_next_plane()

# plt.scatter(x=pix_grid.regular_grid[:,1], y=pix_grid.regular_grid[:,0])
# plt.show()

mapper = adaptive.mapper_from_grid_stack_and_border(grid_stack=source_plane_grid_stack, border=None)


inversion = inv.Inversion(image_1d=lens_data.image_1d, noise_map_1d=lens_data.noise_map_1d,
                          convolver=lens_data.convolver_mapping_matrix, mapper=mapper,
                          regularization=reg.Constant(coefficients=(1.0,)))

# Now lets plot our rectangular mapper with the image.
# mapper_plotters.plot_image_and_mapper(ccd_data=ccd_data, mapper=mapper, mask=mask, should_plot_grid=True)

# Okay, so lets think about the rectangular pixelization. Is it the optimal way to pixelize our source plane? Are there
# features in the source-plane that arn't ideal? How do you think we could do a better job?

# Well, given we're doing a whole tutorial on using a different pixelization to the rectangular grid, you've probably
# guessed that it isn't optimal. Infact, its pretty rubbish, and not a pixelization we'll actually want to model
# many lenses with!

# So what is wrong with the grid? Well, first, lets think about the source reconstruction.
# inversion_plotters.plot_reconstructed_pixelization(inversion=inversion, should_plot_centres=True)

source_galaxy = g.Galaxy(pixelization=adaptive, regularization=reg.Constant(coefficients=(1.0,)))
tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                             image_plane_grid_stack=lens_data.grid_stack)
fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)
lens_fit_plotters.plot_fit_subplot(fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True)


from autolens.model.profiles import light_profiles
from autolens.model.profiles import mass_profiles
from autolens.model.galaxy import galaxy
from autolens.lens import plane
from autolens.data.array import grids
from autolens.lens.plotters import plane_plotters

# We've learnt how to make galaxy objects out of light and mass profiles. Now, we'll use these galaxies to make a
# strong-gravitational lens.
#
# For newcomers to lensing, a strong gravitation lens is a system where two (or more) galaxies align perfectly down
# our line of sight, such that the foreground galaxy's mass (represented as mass profiles) deflects the light
# (represented as light profiles) of the background source galaxy(s).

# When the alignment is just right and the lens is just massive enough, the background source galaxy appears multiple
# times. The schematic below shows a crude drawing of such a system, where two light-rays from the source are bending
# around the lens galaxy and into the observer (light should bend 'smoothly', but drawing this on a keyboard wasn't
# possible - so just pretend the diagonal lines coming from the observer and source are less jagged):

#  Observer                  Image-Plane               Source-Plane
#  (z=0, Earth)               (z = 0.5)                (z = 1.0)
#
#           ----------------------------------------------
#          /                                              \ <---- This is one of the source's light-rays
#         /                      __                       \
#    o   /                      /  \                      __
#    |  /                      /   \                     /  \
#   /\  \                      \   /                     \__/
#        \                     \__/                 Source Galaxy (s)
#         \                Lens Galaxy(s)                /
#           \                                           / <----- And this is its other light-ray
#            ------------------------------------------/

# As an observer, we don't see the source's true appearance (e.g. a round blob of light). Instead, we only observe
# its light after it is deflected and lensed by the foreground galaxy's mass. In this exercise, we'll make a source
# galaxy image whose light has been deflected by a lens galaxy.

# In the schematic above, we used the terms 'Image-Plane' and 'Source-Plane'. In lensing speak, a 'plane' is a
# collection of galaxies at the same redshift (that is, parallel to one another down our line-of-sight). Therefore:

# - If two or more lens galaxies are at the same redshift in the image-plane, they deflect light in the same way.
#   This means we can sum the convergences, potentials and deflection angles of their mass-profiles.

# - If two or more source galaxies are at the same redshift in the source-plane, their light is ray-traced in the
#   same way. Therefore, when determining their lensed images, we can sum the lensed images of each galaxy's light-profiles.

# So, lets do it - lets use the 'plane' module in AutoLens to create a strong lensing system like the one pictured
# above. For simplicity, we'll assume 1 lens galaxy and 1 source galaxy.

# As always, we need grids, where our grids are the coordinates we'll 'trace' from the image-plane to the source-plane
# in the lensing configuration above. Our grid-stack is therefore no longer just a 'grid-stack', but the grid-stack
# representing our image-plane coordinates. Thus, lets name as such.
image_plane_grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(shape=(100, 100), pixel_scale=0.05,
                                                                                  sub_grid_size=2)

# Whereas before we called our galaxy's things like 'galaxy_with_light_profile', lets now refer to them by their role
# in lensing, e.g. 'lens_galaxy' and 'source_galaxy'.
mass_profile = mass_profiles.SphericalIsothermal(centre=(0.0,  0.0), einstein_radius=1.6)
lens_galaxy = galaxy.Galaxy(mass=mass_profile)
light_profile = light_profiles.SphericalSersic(centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0, sersic_index=1.0)
source_galaxy = galaxy.Galaxy(light=light_profile)

# Lets setup our image-plane. This plane takes the lens galaxy we made above and the grid-stack of
# image-plane coordinates.
image_plane = plane.Plane(galaxies=[lens_galaxy], grid_stack=image_plane_grid_stack)

# Up to now, we've kept our galaxies and grids separate, and passed the grid to a galaxy object to compute
# its quantities (e.g. to compute light-profile intensities, we'd write galaxy.intensities_from_grid(grid=grid)).
#
# Plane's combine the galaxies and grids into one object, thus once we've setup a plane there is no longer any need
# have to pass it a grid to compute its quantities. Furthermore, once these quantities are in a plane, they are
# automatically mapped back to their original 2D grid-arrays.
print('deflection-angles of planes regular-grid pixel 1:')
print(image_plane.deflections_y[0,0])
print(image_plane.deflections_x[0,0])
print('deflection-angles of planes regular-grid pixel 2:')
print(image_plane.deflections_y[0,1])
print(image_plane.deflections_x[0,1])

# Plane plotters also don't need grids passed to them anymore - just the plane itself.
plane_plotters.plot_deflections_y(plane=image_plane)
plane_plotters.plot_deflections_x(plane=image_plane)

# Throughout this chapter, we plotted lots of deflection angles. However, if you arn't familiar with strong
# lensing, you probably weren't entirely sure what they are actually used for.
#
# The deflection angles tell us how light is 'lensed' by a lens galaxy. By taking the image-plane coordinates and
# deflection angles, we can subtract the two to determine the source-plane's lensed coordinates, e.g.

# source_plane_coordinates = image_plane_coordinates - image_plane_deflection_angles

# Therefore, we can use our image_plane to 'trace' its grid-stack to the source-plane...
source_plane_grid_stack = image_plane.trace_grid_stack_to_next_plane()

# ... and use this grid-stack to setup the source-plane
source_plane = plane.Plane(galaxies=[source_galaxy], grid_stack=source_plane_grid_stack)

# Lets inspect our regular grids - I bet our source-plane isn't the boring uniform grid we plotted in the first tutorial!
plane_plotters.plot_plane_grid(plane=image_plane, title='Image-plane Grid')
plane_plotters.plot_plane_grid(plane=source_plane, title='Source-plane Grid')

# We can zoom in on the 'centre' of the source-plane (remembering the lens galaxy was centred at (0.0", 0.0")
plane_plotters.plot_plane_grid(plane=source_plane, axis_limits=[-0.1, 0.1, -0.1, 0.1], title='Source-plane Grid')

# We can also plot both planes next to one another, and highlight specific points on the grids. This means we can see
# how different image pixels map to the source-plane (and visa versa).

# (We are inputting the pixel index's into 'points' - the first set of points go from 0 -> 50, which is the top row of
# the image-grid running from the left - as we said it would!)

plane_plotters.plot_image_and_source_plane_subplot(image_plane=image_plane, source_plane=source_plane,
    points=[[range(0,50)], [range(500, 550)],
            [1550, 1650, 1750, 1850, 1950, 2050],
            [8450, 8350, 8250, 8150, 8050, 7950]])

# Clearly, the source-plane's regular-grid is very different to the image-planes! It's not uniform and its centrainly
# not boring!

# We can now ask the question - 'what does our source-galaxy look like in the image-plane'? That is, to us, the
# observer on Earth, how does the source-galaxy appear after lensing?. To do this, we simple trace the source
# galaxy's light 'back' from the source-plane grid-stack.
plane_plotters.plot_image_plane_image(plane=source_plane)

# It's a rather spectacular ring of light, but why is it a ring? Well:

# - Our lens galaxy was centred at (0.0", 0.0").
# - Our source-galaxy was centred at (0.0", 0.0").
# - Our lens galaxy had a spherical mass-profile.
# - Our source-galaxy a spherical light-profile.
#
# Given the perfect symmetry of the system, every path the source's light takes around the lens galaxy is
# radially identical. Thus, nothing else but a ring of light can form!

# This is called an 'Einstein Ring' and its radius is called the 'Einstein Radius', which are both named after the
# man who famously used gravitational lensing to prove his theory of general relativity.

# Finally, because we know our source-galaxy's light profile, we can also plot its 'plane-image'. This image is how
# the source intrinsically appears in the source-plane (e.g. without lensing). This is a useful thing to know, because
# the source-s light is highly magnified, meaning astronomers can study it in a lot more detail than would
# otherwise be possible!
plane_plotters.plot_plane_image(plane=source_plane, plot_grid=True)

# Plotting the grid over the plane image obscures its appearance, which isn't ideal. We can of course tell
# PyAutoLens not to plot the grid.
plane_plotters.plot_plane_image(plane=source_plane, plot_grid=False)

# And, we're done. This is the first tutorial covering strong-lensing and I highly recommend you take a moment
# to really mess about with the code above to see what sort of lensed images you can form. Pay attention to the
# source-plane grid - its appearance can change a lot!
#
# In particular, try:

# 1) Changing the lens galaxy's einstein radius - what happens to the source-plane's image-plane image?

# 2) Change the SphericalIsothermal mass-profile to an EllipticalIsothermal mass-profile and set its axis_ratio to 0.8.
#    What happens to the number of source images?

# 3) As discussed at the beginning, planes can be composed of multiple galaxies. Make an the image-plane with two
#    galaxies and see how multi-galaxy lensing leads to crazy source image-plane images. Also try making a source-plane
#    with two galaxies!

# Finally, if you are a newcomer to strong lensing, it might be worth reading briefly about some strong lensing theory.
# Don't worry about maths, and equations, and anything scary, but you should at least go to Wikipedia to figure out:

# - What a critical line is.

# - What a caustic is.

# - What determines the image multiplicity of the lensed source.


from autofit.tools import path_util
from autolens.data import ccd
from autolens.data.array import grids
from autolens.lens import ray_tracing
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.data.plotters import ccd_plotters
from autolens.lens.plotters import ray_tracing_plotters

from test.simulation import simulation_util

import os


def simulate_image_from_galaxies_and_output_to_fits(data_resolution, data_type, sub_grid_size, lens_galaxies,
                                                    source_galaxies, psf_shape=(51, 51), exposure_time=300.0,
                                                    background_sky_level=0.1):

    pixel_scale = simulation_util.pixel_scale_from_data_resolution(data_resolution=data_resolution)
    shape = simulation_util.shape_from_data_resolution(data_resolution=data_resolution)

    # Simulate a simple Gaussian PSF for the image.
    psf = ccd.PSF.simulate_as_gaussian(shape=psf_shape, sigma=pixel_scale, pixel_scale=pixel_scale)

    # Setup the image-plane grid stack of the CCD array which will be used for generating the image-plane image of the
    # simulated strong lens. A high-res sub-grid is necessary to ensure we fully resolve the central regions of the
    # lens and source galaxy light.
    image_plane_grid_stack = grids.GridStack.grid_stack_for_simulation(shape=shape, pixel_scale=pixel_scale,
                                                                       psf_shape=psf.shape, sub_grid_size=sub_grid_size)

    # Use the input galaxies to setup a tracer, which will generate the image-plane image for the simulated CCD data.
    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=lens_galaxies, source_galaxies=source_galaxies,
                                                 image_plane_grid_stack=image_plane_grid_stack)

    # Simulate the CCD data, remembering that we use a special image-plane image which ensures edge-effects don't
    # degrade our modeling of the telescope optics (e.g. the PSF convolution).
    simulated_ccd_data = ccd.CCDData.simulate(array=tracer.image_plane_image_for_simulation,
                                              pixel_scale=pixel_scale, psf=psf, exposure_time=exposure_time,
                                              background_sky_level=background_sky_level, add_noise=True)

    # Now, lets output this simulated ccd-data to the test/data folder.
    test_path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

    data_path = path_util.make_and_return_path_from_path_and_folder_names(
                path=test_path, folder_names=['data', data_type, data_resolution])

    ccd.output_ccd_data_to_fits(ccd_data=simulated_ccd_data,
                                image_path=data_path + 'image.fits',
                                psf_path=data_path + 'psf.fits',
                                noise_map_path=data_path + 'noise_map.fits',
                                overwrite=True)

    ccd_plotters.plot_ccd_subplot(ccd_data=simulated_ccd_data,
                                  output_filename='ccd_data', output_path=data_path, output_format='png')

    ccd_plotters.plot_ccd_individual(ccd_data=simulated_ccd_data,
                                     should_plot_image=True,
                                     should_plot_noise_map=True,
                                     should_plot_psf=True,
                                     should_plot_signal_to_noise_map=True,
                                     output_path=data_path, output_format='png')

    ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer,  output_filename='tracer', output_path=data_path,
                                                  output_format='png')

    ray_tracing_plotters.plot_ray_tracing_individual(tracer=tracer,
                                                     should_plot_image_plane_image=True,
                                                     should_plot_source_plane=True,
                                                     should_plot_convergence=True,
                                                     should_plot_potential=True,
                                                     should_plot_deflections=True,
                                                     output_path=data_path, output_format='png')

def make_lens_only_dev_vaucouleurs(data_resolutions, sub_grid_size):

    data_type = 'lens_only_dev_vaucouleurs'

    # This lens-only system has a Dev Vaucouleurs spheroid / bulge.

    lens_galaxy = g.Galaxy(bulge=lp.EllipticalDevVaucouleurs(centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0, intensity=0.1,
                                                            effective_radius=1.0))

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(data_type=data_type, data_resolution=data_resolution,
                                                        sub_grid_size=sub_grid_size, lens_galaxies=[lens_galaxy],
                                                        source_galaxies=[g.Galaxy()])

def make_lens_only_bulge_and_disk(data_resolutions, sub_grid_size):

    data_type = 'lens_only_bulge_and_disk'

    # This source-only system has a Dev Vaucouleurs spheroid / bulge and surrounding Exponential envelope

    lens_galaxy = g.Galaxy(bulge=lp.EllipticalDevVaucouleurs(centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0, intensity=0.1,
                                                            effective_radius=1.0),
                           envelope=lp.EllipticalExponential(centre=(0.0, 0.0), axis_ratio=0.7, phi=60.0, intensity=1.0,
                                                             effective_radius=2.0))

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(data_type=data_type, data_resolution=data_resolution,
                                                        sub_grid_size=sub_grid_size, lens_galaxies=[lens_galaxy],
                                                        source_galaxies=[g.Galaxy()])

def make_lens_only_x2_galaxies(data_resolutions, sub_grid_size):

    data_type = 'lens_only_x2_galaxies'

    # This source-only system has two Sersic bulges separated by 2.0"

    lens_galaxy_0 = g.Galaxy(bulge=lp.EllipticalSersic(centre=(-1.0, -1.0), axis_ratio=0.8, phi=0.0, intensity=1.0,
                                                       effective_radius=1.0, sersic_index=3.0))

    lens_galaxy_1 = g.Galaxy(bulge=lp.EllipticalSersic(centre=(1.0, 1.0), axis_ratio=0.8, phi=0.0, intensity=1.0,
                                                       effective_radius=1.0, sersic_index=3.0))

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(data_type=data_type, data_resolution=data_resolution,
                                                        sub_grid_size=sub_grid_size,
                                                        lens_galaxies=[lens_galaxy_0, lens_galaxy_1],
                                                        source_galaxies=[g.Galaxy()])

def make_no_lens_light_and_source_smooth(data_resolutions, sub_grid_size):

    data_type = 'no_lens_light_and_source_smooth'

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.6,
                                                        axis_ratio=0.7, phi=45.0))

    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=60.0,
                                                       intensity=0.4, effective_radius=0.5, sersic_index=1.0))

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(data_type=data_type, data_resolution=data_resolution,
                                                        sub_grid_size=sub_grid_size,
                                                        lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy])

def make_no_lens_light_and_source_cuspy(data_resolutions, sub_grid_size):

    data_type = 'no_lens_light_and_source_cuspy'

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.6,
                                                        axis_ratio=0.7, phi=45.0))

    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=60.0,
                                                       intensity=0.1, effective_radius=0.5, sersic_index=3.0))

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(data_type=data_type, data_resolution=data_resolution,
                                                        sub_grid_size=sub_grid_size,
                                                        lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy])

def make_no_lens_light_and_source_smooth_offset_centre(data_resolutions, sub_grid_size):

    data_type = 'no_lens_light_and_source_smooth_offset_centre'

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(1.0, 1.0), einstein_radius=1.6,
                                                        axis_ratio=0.7, phi=45.0))

    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(1.0, 1.0), axis_ratio=0.8, phi=60.0,
                                                       intensity=0.4, effective_radius=0.5, sersic_index=1.0))

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(data_type=data_type, data_resolution=data_resolution,
                                                        sub_grid_size=sub_grid_size,
                                                        lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy])

def make_lens_and_source_smooth(data_resolutions, sub_grid_size):

    data_type = 'lens_and_source_smooth'

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = g.Galaxy(ight=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0,
                                                 intensity=0.5, effective_radius=0.8, sersic_index=4.0),
                           mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.6,
                                                        axis_ratio=0.7, phi=45.0))

    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=60.0,
                                                       intensity=0.4, effective_radius=0.5, sersic_index=1.0))

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(data_type=data_type, data_resolution=data_resolution,
                                                        sub_grid_size=sub_grid_size,
                                                        lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy])

def make_lens_and_source_cuspy(data_resolutions, sub_grid_size):

    data_type = 'lens_and_source_cuspy'

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = g.Galaxy(ight=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0,
                                                 intensity=0.5, effective_radius=0.8, sersic_index=4.0),
                           mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.6,
                                                        axis_ratio=0.7, phi=45.0))

    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=60.0,
                                                       intensity=0.1, effective_radius=0.5, sersic_index=3.0))

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(data_type=data_type, data_resolution=data_resolution,
                                                        sub_grid_size=sub_grid_size,
                                                        lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy])


from autofit.tools import path_util
from autofit.optimize import non_linear as nl
from autofit.mapper import prior
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline
from autolens.pipeline import tagging as tag
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp

# In this pipeline, we'll demonstrate binning up - which allows us to fit a binned up version of an image in a phse of 
# the pipeline. In this example, we will perform an initial analysis on an image binned up from a pixel scale of 0.05" 
# to 0.01", which gives significant speed-up in run timme, and then refine the model in a second phase at the input
# resolution.

# Whilst bin up factors can be manually specified in the pipeline, in this example we will make the bin up factor
# an input parameter of the pipeline. This means we can run the pipeline with different binning up factors for different
# runners.

# We will also use phase tagging to ensure phases which use binned up data have a tag in their path, so it is clear
# that a phase uses this feature.

# We'll perform a basic analysis which fits a lensed source galaxy using a parametric light profile where
# the lens's light is omitted. This pipeline uses two phases:

# Phase 1:

# Description: Initializes the lens mass model and source light profile using x1 source with a bin up factor of x2.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: Uses a bin up factor of x2

# Phase 1:

# Description: Fits the lens and source model using unbinned data.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: Lens mass (variable -> phase 1), source light (variable -> phase 1)
# Notes: No binning up.

def make_pipeline(phase_folders=None, bin_up_factor=2):

    pipeline_name = 'pipeline_binning_up'

    # This tag is 'added' to the phase path, to make it clear what binning up is used. The bin_up_tag and phase
    # name are shown for 3 example bin up factors:

    # bin_up_factor=1 -> phase_path=phase_name
    # bin_up_factor=2 -> phase_path=phase_name_bin_up_factor_2
    # bin_up_factor=3 -> phase_path=phase_name_bin_up_factor_3

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/'
    phase_folders = path_util.phase_folders_from_phase_folders_and_pipeline_name(phase_folders=phase_folders,
                                                                                pipeline_name=pipeline_name)

    # As there is no lens light component, we can use an annular mask throughout this pipeline which removes the
    # central regions of the image.

    def mask_function(image):
        return msk.Mask.circular_annular(shape=image.shape, pixel_scale=image.pixel_scale,
                                         inner_radius_arcsec=0.2, outer_radius_arcsec=3.3)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy (y,x) centre such that we assume the image is centred around the lens galaxy.

    class LensSourceX1Phase(ph.LensSourcePlanePhase):

        def pass_priors(self, results):

            self.lens_galaxies.lens.mass.centre_0 = prior.GaussianPrior(mean=0.0, sigma=0.1)
            self.lens_galaxies.lens.mass.centre_1 = prior.GaussianPrior(mean=0.0, sigma=0.1)

    phase1 = LensSourceX1Phase(phase_name='phase_1_x1_source', phase_folders=phase_folders,
                               phase_tagging=True,
                               lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal,
                                                                      shear=mp.ExternalShear)),
                               source_galaxies=dict(source_0=gm.GalaxyModel(light=lp.EllipticalSersic)),
                               bin_up_factor=bin_up_factor,
                               mask_function=mask_function, optimizer_class=nl.MultiNest)

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # In phase 2, we will fit the lens galaxy's mass and two source galaxies, where we:

    # 1) Initialize the priors on the lens galaxy using the results of phase 1.
    # 2) Initialize the priors on the first source galaxy using the results of phase 1.

    class LensSourceX2Phase(ph.LensSourcePlanePhase):

        def pass_priors(self, results):

            self.lens_galaxies.lens = results.from_phase('phase_1_x1_source').variable.lens
            self.source_galaxies.source_0 = results.from_phase('phase_1_x1_source').variable.source_0

    phase2 = LensSourceX2Phase(phase_name='phase_2_x2_source', phase_folders=phase_folders,
                               phase_tagging=True,
                               lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal,
                                                                      shear=mp.ExternalShear)),
                               source_galaxies=dict(source_0=gm.GalaxyModel(light=lp.EllipticalSersic),
                                                      source_1=gm.GalaxyModel(light=lp.EllipticalSersic)),
                               optimizer_class=nl.MultiNest, mask_function=mask_function)

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.3

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2)


from autofit.tools import path_util
from autofit.optimize import non_linear as nl
from autofit.mapper import prior
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp

import os

############## THIS PIPELINE IS A WORK IN PROGRESS AND I DO NOT RECOMMEND YOU TRY TO USE IT TO DO LENS MODELING ########

# In this pipeline, we'll perform an advanced analysis which fits a lens galaxy with three surrounding line-of-sight \
# galaxies (which are at different redshifts and thus define a multi-plane stronog lens configuration). The source \
# galaxy will be modeled using a parametric light profile in the initial phases, but switch to an inversion in the \
# later phases.

# For efficiency, we will subtract the lens galaxy's light and line-of-sight galaxies light, and then only fit a
# foreground subtracted image. This means we can use an annular mask tailored to the source galaxy's light, which we
# have setup already using the 'tools/mask_maker.py' scripts.

# This leads to a 5 phase pipeline:

# Phase 1) Fit and subtract the light profile of the lens galaxy (elliptical Sersic) and each line-of-sight
#          galaxy (spherical Sersic).

# Phase 2) Use this lens subtracted image to fit the lens galaxy's mass (SIE) and source galaxy's light (Sersic),
#          thereby omitting multi-plane ray-tracing.

# Phase 3) Initialize the resolution and regularization coefficient of the inversion using the best-fit lens model from
#          phase 2.

# Phase 4) Refit the lens galaxy's light and mass models using an inversion (with priors from the above phases). The
#          mass profile of the 3 line-of-sight galaxies is also included (as SIS profiles), but a single lens plane
#          is assumed.

# Phase 5) Fit the lens galaxy, line-of-sight galaxies and source-galaxy using multi-plane ray-tracing, where the
#          redshift of each line-of-sight galaxy is included in the non-linear search as a free parameter. This phase
#          uses priors initialized from phase 4.

# Phase 5 risks the inversion falling into the systematic solutions where the mass of the lens model is too high,
# resulting in the inversion reconstructing the image as a demagnified version of itself. This is because the \
# line-of-sight halos each have a mass, which during the modeling can go to very high values.

# For this reason, we include positions (drawn using the 'tools/positions.py' script) to prevent these solutions from
# existing in parameter space.

def make_pipeline(phase_folders=None):

    pipeline_name = 'pipeline_multi_plane'

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/'
    phase_folders = path_util.phase_folders_from_phase_folders_and_pipeline_name(phase_folders=phase_folders,
                                                                                pipeline_name=pipeline_name)

    # In phase 1, we will:

    # 1) Subtract the light of the main lens galaxy (located at (0.0", 0.0")) and the light of each line-of-sight
    # galaxy (located at (4.0", 4.0"), (3.6", -5.3") and (-3.1", -2.4"))

    class LensPhase(ph.LensPlanePhase):

        def pass_priors(self, results):

            self.lens_galaxies.lens.light.centre_0 = prior.GaussianPrior(mean=0.0, sigma=0.1)
            self.lens_galaxies.lens.light.centre_1 = prior.GaussianPrior(mean=0.0, sigma=0.1)
            self.lens_galaxies.los0.light.centre_0 = prior.GaussianPrior(mean=4.0, sigma=0.1)
            self.lens_galaxies.los0.light.centre_1 = prior.GaussianPrior(mean=4.0, sigma=0.1)
            self.lens_galaxies.los1.light.centre_0 = prior.GaussianPrior(mean=3.6, sigma=0.1)
            self.lens_galaxies.los1.light.centre_1 = prior.GaussianPrior(mean=-5.3, sigma=0.1)
            self.lens_galaxies.los2.light.centre_0 = prior.GaussianPrior(mean=-3.1, sigma=0.1)
            self.lens_galaxies.los2.light.centre_1 = prior.GaussianPrior(mean=-2.4, sigma=0.1)

    phase1 = LensPhase(phase_name='phase_1_light_subtraction', phase_folders=phase_folders,
                       lens_galaxies=dict(lens=gm.GalaxyModel(light=lp.EllipticalSersic),
                                           los_0=gm.GalaxyModel(light=lp.SphericalSersic),
                                           los_1=gm.GalaxyModel(light=lp.SphericalSersic),
                                           los_2=gm.GalaxyModel(light=lp.SphericalSersic)),
                       optimizer_class=nl.MultiNest)

    # Customize MultiNest so it runs fast
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2
    phase1.optimizer.const_efficiency_mode = True

    # In phase 2, we will:

    # 1) Fit this foreground subtracted image, using an SIE+Shear mass model and Sersic source.
    # 2) Use the input positions to resample inaccurate mass models.

    class LensSubtractedPhase(ph.LensSourcePlanePhase):

        def modify_image(self, image, results):
            return image - results.from_phase('phase_1_light_subtraction').unmasked_lens_plane_model_image

        def pass_priors(self, results):

            self.lens_galaxies.lens.mass.centre_0 = prior.GaussianPrior(mean=0.0, sigma=0.1)
            self.lens_galaxies.lens.mass.centre_1 = prior.GaussianPrior(mean=0.0, sigma=0.1)

    phase2 = LensSubtractedPhase(phase_name='phase_2_source_parametric', phase_folders=phase_folders,
                                 lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                                 source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalSersic)),
                                 positions_threshold=0.3,
                                 optimizer_class=nl.MultiNest)

    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.2
    phase2.optimizer.const_efficiency_mode = True

    # In phase 3, we will:

    #  1) Fit the foreground subtracted image using a source-inversion instead of parametric source, using lens galaxy
    #     priors from phase 2.

    class InversionPhase(ph.LensSourcePlanePhase):

        def modify_image(self, image, results):
            return image - results.from_phase('phase_1_light_subtraction').unmasked_lens_plane_model_image

        def pass_priors(self, results):

            self.lens_galaxies.lens = results[1].constant.lens

            self.source_galaxies.source.pixelization.shape_0 = prior.UniformPrior(lower_limit=20.0, upper_limit=45.0)
            self.source_galaxies.source.pixelization.shape_1 = prior.UniformPrior(lower_limit=20.0, upper_limit=45.0)

    phase3 = InversionPhase(phase_name='phase_3_inversion_init', phase_folders=phase_folders,
                            lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                            source_galaxies=dict(source=gm.GalaxyModel(pixelization=pix.AdaptiveMagnification,
                                                                      regularization=reg.Constant)),
                            optimizer_class=nl.MultiNest)

    # Customize MultiNest so it runs fast
    phase3.optimizer.n_live_points = 10
    phase3.optimizer.sampling_efficiency = 0.5
    phase3.optimizer.const_efficiency_mode = True

    # In phase 4, we will:

    #  1) Fit the foreground subtracted image using a source-inversion with parameters initialized from phase 3.
    #  2) Initialize the lens galaxy priors using the results of phase 2, and include each line-of-sight galaxy's mass
    #     contribution (keeping its centre fixed to the light profile centred inferred in phase 1). This will assume
    #     a single lens plane for all line-of-sight galaxies.
    #  3) Use the input positions to resample inaccurate mass models.

    class SingleLensPlanePhase(ph.LensSourcePlanePhase):

        def modify_image(self, image, results):
            return image - results.from_phase('phase_1_light_subtraction').unmasked_lens_plane_model_image

        def pass_priors(self, results):

            self.lens_galaxies.lens.mass = results.from_phase('phase_3_inversion_init').variable.lens.mass

            self.lens_galaxies.los_0.mass.centre_0 = results[0].constant.los_0.light.centre_0
            self.lens_galaxies.los_0.mass.centre_1 = results[0].constant.los_0.light.centre_1
            self.lens_galaxies.los_1.mass.centre_0 = results[0].constant.los_1.light.centre_0
            self.lens_galaxies.los_1.mass.centre_1 = results[0].constant.los_1.light.centre_1
            self.lens_galaxies.los_2.mass.centre_0 = results[0].constant.los_2.light.centre_0
            self.lens_galaxies.los_2.mass.centre_1 = results[0].constant.los_2.light.centre_1

            self.source_galaxies.source = results.from_phase('phase_3_inversion_init').variable.source

    phase4 = SingleLensPlanePhase(phase_name='phase_4_single_plane', phase_folders=phase_folders,
                                  lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal),
                                                     los_0=gm.GalaxyModel(mass=mp.SphericalIsothermal),
                                                     los_1=gm.GalaxyModel(mass=mp.SphericalIsothermal),
                                                     los_2=gm.GalaxyModel(mass=mp.SphericalIsothermal)),
                                  source_galaxies=dict(source=gm.GalaxyModel(pixelization=pix.AdaptiveMagnification,
                                                                             regularization=reg.Constant)),
                                  positions_threshold=0.3,
                                  optimizer_class=nl.MultiNest)

    # Customize MultiNest so it runs fast
    phase4.optimizer.n_live_points = 60
    phase4.optimizer.sampling_efficiency = 0.2
    phase4.optimizer.const_efficiency_mode = True


    # In phase 5, we will fit the foreground subtracted image using a multi-plane ray tracer. This means that the
    # redshift of each line-of-sight galaxy is included as a free parameter (we assume the lens and source redshifts
    # are known).

    class MultiPlanePhase(ph.MultiPlanePhase):

        def modify_image(self, image, results):
            return image - results.from_phase('phase_1_light_subtraction').unmasked_lens_plane_model_image

        def pass_priors(self, results):

            self.galaxies.lens = results.from_phase('phase_4_single_plane').variable.lens

            self.galaxies.los_0.mass = results.from_phase('phase_4_single_plane').variable.los_0.mass
            self.galaxies.los_1.mass = results.from_phase('phase_4_single_plane').variable.los_1.mass
            self.galaxies.los_2.mass = results.from_phase('phase_4_single_plane').variable.los_2.mass

            self.galaxies.source = results.from_phase('phase_4_single_plane').variable.source

    phase5 = MultiPlanePhase(phase_name='phase_5_multi_plane', phase_folders=phase_folders,
                             galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal),
                             los_0=gm.GalaxyModel(mass=mp.SphericalIsothermal, variable_redshift=True),
                             los_1=gm.GalaxyModel(mass=mp.SphericalIsothermal, variable_redshift=True),
                             los_2=gm.GalaxyModel(mass=mp.SphericalIsothermal, variable_redshift=True),
                             source=gm.GalaxyModel(pixelization=pix.AdaptiveMagnification,
                                                   regularization=reg.Constant)),
                             positions_threshold=0.3,
                             optimizer_class=nl.MultiNest)

    # Customize MultiNest so it runs fast
    phase5.optimizer.n_live_points = 60
    phase5.optimizer.sampling_efficiency = 0.2
    phase5.optimizer.const_efficiency_mode = True

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2, phase3, phase4, phase5)


from autolens.pipeline import tagging

class TestPhaseTag:

    def test__mixture_of_values(self):

        phase_tag = tagging.phase_tag_from_phase_settings(sub_grid_size=2,
                                                          bin_up_factor=None,
                                                          image_psf_shape=None,
                                                          inversion_psf_shape=None,
                                                          inner_mask_radii=0.3,
                                                          positions_threshold=2.0,
                                                          interp_pixel_scale=None)


        assert phase_tag == '_sub_2_pos_2.00_inner_mask_0.30'

        phase_tag = tagging.phase_tag_from_phase_settings(sub_grid_size=1,
                                                          bin_up_factor=3,
                                                          image_psf_shape=(2, 2),
                                                          inversion_psf_shape=(3,3),
                                                          inner_mask_radii=None,
                                                          positions_threshold=None,
                                                          interp_pixel_scale=0.2)

        assert phase_tag == '_sub_1_bin_up_3_image_psf_2x2_inversion_psf_3x3_interp_0.200'

class TestTaggers:

    def test__positions_threshold_tagger(self):

        tag = tagging.positions_threshold_tag_from_positions_threshold(positions_threshold=None)
        assert tag == ''
        tag = tagging.positions_threshold_tag_from_positions_threshold(positions_threshold=1.0)
        assert tag == '_pos_1.00'
        tag = tagging.positions_threshold_tag_from_positions_threshold(positions_threshold=2.56)
        assert tag == '_pos_2.56'

    def test__inner_circular_mask_radii_tagger(self):

        tag = tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(inner_mask_radii=None)
        assert tag == ''
        tag = tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(inner_mask_radii=0.2)
        print(tag)
        assert tag == '_inner_mask_0.20'
        tag = tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(inner_mask_radii=3)
        assert tag == '_inner_mask_3.00'

    def test__sub_grid_size_tagger(self):

        tag = tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=1)
        assert tag == '_sub_1'
        tag = tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=2)
        assert tag == '_sub_2'
        tag = tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=4)
        assert tag == '_sub_4'

    def test__bin_up_factor_tagger(self):

        tag = tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=None)
        assert tag == ''
        tag = tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=1)
        assert tag == ''
        tag = tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=2)
        assert tag == '_bin_up_2'
        tag = tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=3)
        assert tag == '_bin_up_3'

    def test__image_psf_shape_tagger(self):

        tag = tagging.image_psf_shape_tag_from_image_psf_shape(image_psf_shape=None)
        assert tag == ''
        tag = tagging.image_psf_shape_tag_from_image_psf_shape(image_psf_shape=(2,2))
        print(tag.strip())
        assert tag == '_image_psf_2x2'
        tag = tagging.image_psf_shape_tag_from_image_psf_shape(image_psf_shape=(3,4))
        assert tag == '_image_psf_3x4'

    def test__inversion_psf_shape_tagger(self):

        tag = tagging.inversion_psf_shape_tag_from_inversion_psf_shape(inversion_psf_shape=None)
        assert tag == ''
        tag = tagging.inversion_psf_shape_tag_from_inversion_psf_shape(inversion_psf_shape=(2,2))
        print(tag.strip())
        assert tag == '_inversion_psf_2x2'
        tag = tagging.inversion_psf_shape_tag_from_inversion_psf_shape(inversion_psf_shape=(3,4))
        assert tag == '_inversion_psf_3x4'

    def test__interp_pixel_scale_tagger(self):

        tag = tagging.interp_pixel_scale_tag_from_interp_pixel_scale(interp_pixel_scale=None)
        assert tag == ''
        tag = tagging.interp_pixel_scale_tag_from_interp_pixel_scale(interp_pixel_scale=0.5)
        assert tag == '_interp_0.500'
        tag = tagging.interp_pixel_scale_tag_from_interp_pixel_scale(interp_pixel_scale=0.25)
        assert tag == '_interp_0.250'
        tag = tagging.interp_pixel_scale_tag_from_interp_pixel_scale(interp_pixel_scale=0.234)
        assert tag == '_interp_0.234'


from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy
from autolens.lens import ray_tracing
from autolens.data.array import grids
from autolens.lens.plotters import plane_plotters
from autolens.lens.plotters import ray_tracing_plotters

# In this example, we'll use the 'ray-tracing' module, to setup the same lens-plane + source-plane strong
# lens configuration as the previous tutorial, but with a lot less lines of code!

# Let use the same grid-stack we've all grown to know and love by now!
image_plane_grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(shape=(100, 100), pixel_scale=0.1,
                                                                                  sub_grid_size=2)

# subhalo_0 = galaxy.Galaxy(mass=mp.SphericalNFW(centre=(-1.0, -1.0), kappa_s=1.0, scale_radius=0.5), redshift=0.5)
# subhalo_1 = galaxy.Galaxy(mass=mp.SphericalNFW(centre=(1.0, 1.0), kappa_s=1.0, scale_radius=0.5), redshift=0.4)

lens_galaxy = galaxy.Galaxy(mass=mp.EllipticalIsothermal(centre=(3.0, 3.0), axis_ratio=0.8, phi=45.0,
                                                         einstein_radius=1.0), redshift=0.5)

sheet_galaxy = galaxy.Galaxy(sheet=mp.MassSheet(centre=(0.0, 0.0), kappa=0.0), redshift=0.2)

source_galaxy = galaxy.Galaxy(light=lp.SphericalSersic(centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0,
                                                      sersic_index=3.0), redshift=1.0)

tracer = ray_tracing.TracerMultiPlanes(galaxies=[lens_galaxy, sheet_galaxy, source_galaxy],
                                       image_plane_grid_stack=image_plane_grid_stack)

ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer)

tracer = ray_tracing.TracerMultiPlanesSliced(lens_galaxies=[lens_galaxy],
                                             line_of_sight_galaxies=[sheet_galaxy],
                                             source_galaxies=[source_galaxy], planes_between_lenses=[2, 2],
                                             image_plane_grid_stack=image_plane_grid_stack)

ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer)


from autofit.tools import path_util
from autofit import conf
from autolens.data import ccd
from autolens.data.plotters import ccd_plotters

import os

# Welcome to the pipeline runner. This tool allows you to load data on strong lenses, and pass it to pipelines for a
# strong lens analysis. To show you around, we'll load up some example data and run it through some of the example
# pipelines that come distributed with PyAutoLens, in the 'autolens_workspace/pipelines/simple' folder.
#
# In this  runner, our analysis assumes there is no lens light component that requires modeling and subtraction.

# Why is this folder called 'simple'? Well, we actually recommend you break pipelines up. Details of how to do this
# can be found in the 'autolens_workspace/runners/advanced' folder, but its a bit conceptually difficult. So, to
# familiarize yourself with PyAutoLens, I'd stick to the simple pipelines for now!

# You should also checkout the 'autolens_workspace/runners/features' folder. Here, you'll find features that are
# available in pipelines, that allow you to customize the analysis.

# This runner is supplied as both this Python script and a Juypter notebook. Its up to you which you use - I personally
# prefer the python script as provided you keep it relatively small, its quick and easy to comment out different lens
# names and pipelines to set off different analyses. However, notebooks are a tidier way to manage visualization - so
# feel free to use notebooks. Or, use both for a bit, and decide your favourite!

# Setup the path to the workspace, using a relative directory name.
workspace_path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output path.
conf.instance = conf.Config(config_path=workspace_path + 'config', output_path=workspace_path + 'output')

# It is convenient to specify the lens name as a string, so that if the pipeline is applied to multiple images we \
# don't have to change all of the path entries in the load_ccd_data_from_fits function below.

data_type = 'example'
data_name = 'lens_mass_and_x1_source' # An example simulated image without any lens light and a source galaxy.
pixel_scale = 0.1

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/example/lens_light_mass_and_x1_source/'
data_path = path_util.make_and_return_path_from_path_and_folder_names(path=workspace_path,
                                                                      folder_names=['data', data_type, data_name])

ccd_data = ccd.load_ccd_data_from_fits(image_path=data_path + 'image.fits',
                                       psf_path=data_path + 'psf.fits',
                                       noise_map_path=data_path + 'noise_map.fits',
                                       pixel_scale=pixel_scale)

ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# Running a pipeline is easy, we simply import it from the pipelines folder and pass the lens data to its run function.
# Below, we'll use a 3 phase example pipeline to fit the data with a mass model and pixelized source reconstruction.
# Checkout _workspace/pipelines/examples/lens_sie_shear_source_inversion.py_' for a full description of
# the pipeline.

from workspace.pipelines.simple import lens_sie_source_inversion
pipeline = lens_sie_source_inversion.make_pipeline(phase_folders=[data_type, data_name])
pipeline.run(data=ccd_data)

# Another example pipeline is shown below, which fits an image generated using two source galaxies with a
# pipeline that fits two source galaxies. The source light is modeled as parametric Sersic functions.

# data_type = 'example'
# data_name = 'lens_mass_and_x2_source' # An example simulated image without any lens light and two source galaxies.
# pixel_scale = 0.1
#
# data_path = path_util.make_and_return_path_from_path_and_folder_names(path=workspace_path,
#                                                                       folder_names=['data', data_type, data_name])
#
# ccd_data = ccd.load_ccd_data_from_fits(image_path=data_path + 'image.fits',
#                                        psf_path=data_path + 'psf.fits',
#                                        noise_map_path=data_path + 'noise_map.fits',
#                                        pixel_scale=pixel_scale)
#
# ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# from workspace.pipelines.simple import lens_sie_source_x2_sersic
# pipeline = lens_sie_source_x2_sersic.make_pipeline(phase_folders=[data_type, data_name])
# pipeline.run(data=ccd_data)


import os

from autofit import conf
from autofit.optimize import non_linear as nl
from autofit.mapper import prior
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from test.integration import integration_util
from test.simulation import simulation_util

test_type = 'sensitivity'
test_name = "sensitivity_profile_via_multinest"

test_path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + 'output/'
config_path = test_path + 'config'
conf.instance = conf.Config(config_path=config_path, output_path=output_path)


def pipeline():

    integration_util.reset_paths(test_name=test_name, output_path=output_path)
    ccd_data = simulation_util.load_test_ccd_data(data_type='no_lens_light_and_source_smooth', data_resolution='Euclid')
    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(data=ccd_data)


def make_pipeline(test_name):

    class SensitivePhase(ph.SensitivityPhase):

        def pass_priors(self, results):

            self.lens_galaxies.lens.mass.centre_0 = 0.0
            self.lens_galaxies.lens.mass.centre_1 = 0.0
            self.lens_galaxies.lens.mass.einstein_radius = 1.6

            self.source_galaxies.source.light.centre_0 = 0.0
            self.source_galaxies.source.light.centre_1 = 0.0
            self.source_galaxies.source.light.intensity = 1.0
            self.source_galaxies.source.light.effective_radius = 0.5
            self.source_galaxies.source.light.sersic_index = 1.0

            self.sensitive_galaxies.subhalo.mass.centre_0 = prior.GaussianPrior(mean=0.0, sigma=1.0)
            self.sensitive_galaxies.subhalo.mass.centre_1 = prior.GaussianPrior(mean=0.0, sigma=1.0)
            self.sensitive_galaxies.subhalo.mass.kappa_s = 0.1
            self.sensitive_galaxies.subhalo.mass.scale_radius = 5.0

    phase1 = SensitivePhase(phase_name="phase1", phase_folders=[test_type, test_name],
                            lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.SphericalIsothermal)),
                            source_galaxies=dict(source=gm.GalaxyModel(light=lp.SphericalSersic)),
                            sensitive_galaxies=dict(subhalo=gm.GalaxyModel(mass=mp.SphericalNFW)),
                            optimizer_class=nl.MultiNest)

    phase1.optimizer.const_efficiency_mode = True

    return pl.PipelineImaging(test_name, phase1)


if __name__ == "__main__":
    pipeline()



from autofit import conf
from autolens.data import ccd
from autolens.data.array import mask as msk
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy as g
from autolens.lens import lens_data as ld
from autolens.lens import ray_tracing
from autolens.lens import lens_fit
from autolens.data.plotters import ccd_plotters
from autolens.lens.plotters import lens_fit_plotters

import os

# So far, we've not paid much attention to the source galaxy's morphology. We've assumed its a single-component
# exponential profile, which is a fairly crude assumption. A quick look at any image of a real galaxy reveals a wealth
# of different structures that could be present - bulges, disks, bars, star-forming knots and so on.
# Furthermore, there could be more than one source-galaxy!

# In this example, we'll explore how far we can get trying to_fit a complex source using a pipeline. Fitting complex
# source's is an exercise in diminishing returns. Each component we add to our source model brings with it an extra 5-7,
# parameters. If there are 4 components, or multiple galaxies, we're quickly entering the somewhat nasty regime of
# 30-40+ parameters in our non-linear search. Even with a pipeline, thats a lot of parameters to try and fit!

# Lets setup the path to the workspace, as per usual.
workspace_path = '{}/../../../'.format(os.path.dirname(os.path.realpath(__file__)))
conf.instance = conf.Config(config_path=workspace_path + 'config', output_path=workspace_path + 'output')

# This function simulates an image with a complex source.
def simulate():

    from autolens.data.array import grids
    from autolens.model.galaxy import galaxy as g
    from autolens.lens import ray_tracing

    psf = ccd.PSF.simulate_as_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    image_plane_grid_stack = grids.GridStack.grid_stack_for_simulation(shape=(180, 180), pixel_scale=0.05,
                                                                       psf_shape=(11, 11))

    lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal( centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0,
                                                         einstein_radius=1.6))
    source_galaxy_0 = g.Galaxy(light=lp.EllipticalSersic(centre=(0.1, 0.1), axis_ratio=0.8, phi=90.0, intensity=0.2,
                                                         effective_radius=1.0, sersic_index=1.5))
    source_galaxy_1 = g.Galaxy(light=lp.EllipticalSersic(centre=(-0.25, 0.25), axis_ratio=0.7, phi=45.0, intensity=0.1,
                                                         effective_radius=0.2, sersic_index=3.0))
    source_galaxy_2 = g.Galaxy(light=lp.EllipticalSersic(centre=(0.45, -0.35), axis_ratio=0.6, phi=90.0, intensity=0.03,
                                                         effective_radius=0.3, sersic_index=3.5))
    source_galaxy_3 = g.Galaxy(light=lp.EllipticalSersic(centre=(-0.05, -0.0), axis_ratio=0.9, phi=140.0, intensity=0.03,
                                                         effective_radius=0.1, sersic_index=4.0))

    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy],
                                                 source_galaxies=[source_galaxy_0, source_galaxy_1,
                                                                  source_galaxy_2, source_galaxy_3],
                                                 image_plane_grid_stack=image_plane_grid_stack)

    return ccd.CCDData.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.05,
                               exposure_time=300.0, psf=psf, background_sky_level=0.1, add_noise=True)

# Lets simulate the image.
ccd_data = simulate()
ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# Yep, that's a pretty complex source. There are clearly more than 4 peaks of light - I wouldn't like to guess whether
# there is one or two sources (or more). You'll also notice I omitted the lens galaxy's light for this system, this is
# to keep the number of parameters down and the phases running fast, but we wouldn't get such a luxury in a real galaxy.

# Again, before we checkout the pipeline, lets import it, and get it running.
from workspace.howtolens.chapter_3_pipelines import tutorial_3_pipeline_complex_source

pipeline_complex_source = tutorial_3_pipeline_complex_source.make_pipeline(
    phase_folders=['howtolens', 'c3_t3_complex_source'])
pipeline_complex_source.run(data=ccd_data)

# Okay, so with 4 sources, we still couldn't get a good a fit to the source that didn't leave residuals. The thing is,
# I did infact simulate the lens with 4 sources. This means that there is a 'perfect fit', somewhere in that parameter
# space, that we unfortunately missed using the pipeline above.
#
# Lets confirm this, by manually fitting the ccd data with the true input model.

lens_data = ld.LensData(ccd_data=ccd_data, mask=msk.Mask.circular(shape=ccd_data.shape,
                                                                  pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.0))

lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0,
                                                    einstein_radius=1.6))
source_galaxy_0 = g.Galaxy(light=lp.EllipticalSersic(centre=(0.1, 0.1), axis_ratio=0.8, phi=90.0, intensity=0.2,
                                                     effective_radius=1.0, sersic_index=1.5))
source_galaxy_1 = g.Galaxy(light=lp.EllipticalSersic(centre=(-0.25, 0.25), axis_ratio=0.7, phi=45.0, intensity=0.1,
                                                     effective_radius=0.2, sersic_index=3.0))
source_galaxy_2 = g.Galaxy(light=lp.EllipticalSersic(centre=(0.45, -0.35), axis_ratio=0.6, phi=90.0, intensity=0.03,
                                                     effective_radius=0.3, sersic_index=3.5))
source_galaxy_3 = g.Galaxy(light=lp.EllipticalSersic(centre=(-0.05, -0.0), axis_ratio=0.9, phi=140.0, intensity=0.03,
                                                     effective_radius=0.1, sersic_index=4.0))

tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy],
                                             source_galaxies=[source_galaxy_0, source_galaxy_1,
                                                              source_galaxy_2, source_galaxy_3],
                                             image_plane_grid_stack=lens_data.grid_stack)

true_fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

lens_fit_plotters.plot_fit_subplot(fit=true_fit)

# And indeed, we see far improved residuals, chi-squareds, etc.

# The morale of this story is that, if the source morphology is complex, there is no way we can build a pipeline to
# fit it. For this tutorial, this was true even though our source model could actually fit the data perfectly. For
# real lenses, the source will be *even more complex* and there is even less hope of getting a good fit :(

# But fear not, PyAutoLens has you covered. In chapter 4, we'll introduce a completely new way to model the source
# galaxy, which addresses the problem faced here. But before that, in the next tutorial we'll discuss how we actually
# pass priors in a pipeline.


import shutil
import os
import numpy as np
import pytest

from autolens.data.array.util import mask_util as util
from autolens.data.array.util import mapping_util
from autolens.data.array import mask as msk
from autolens.data.array.util import grid_util

test_data_dir = "{}/../../test_files/array/".format(os.path.dirname(os.path.realpath(__file__)))


class TestMask:

    def test__constructor(self):

        mask = np.array([[True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True]])

        mask = msk.Mask(mask, pixel_scale=1)

        assert (mask == np.array([[True, True, True, True],
                                 [True, False, False, True],
                                 [True, True, True, True]])).all()
        assert mask.pixel_scale == 1.0
        assert mask.central_pixel_coordinates == (1.0, 1.5)
        assert mask.shape == (3, 4)
        assert mask.shape_arcsec == (3.0, 4.0)

    def test__array_finalize__masks_pass_attributes(self):

        mask = np.array([[True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True]])

        mask = msk.Mask(array=mask, pixel_scale=1.0)

        mask_new = mask + mask

        assert mask_new.pixel_scale == 1.0
        assert mask_new.origin == (0.0, 0.0)
        assert mask_new.centre == (0.0, 0.0)

    def test__centring__adapts_to_max_and_min_of_mask(self):

        mask = np.array([[True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True]])

        mask = msk.Mask(mask, pixel_scale=1.0)

        assert mask.centre == (0.0, 0.0)

        mask = np.array([[True, True, True, True],
                         [True, False, False, False],
                         [True, True, True, True]])

        mask = msk.Mask(mask, pixel_scale=1.0)

        assert mask.centre == (0.0, 0.5)

        mask = np.array([[True, True, False, True],
                         [True, False, False, True],
                         [True, True, True, True]])

        mask = msk.Mask(mask, pixel_scale=1.0)

        assert mask.centre == (0.5, 0.0)

        mask = np.array([[True, True, True, True],
                         [False, False, False, True],
                         [True, True, True, True]])

        mask = msk.Mask(mask, pixel_scale=1.0)

        assert mask.centre == (0.0, -0.5)

        mask = np.array([[True, True, True, True],
                         [True, False, False, True],
                         [True, False, True, True]])

        mask = msk.Mask(mask, pixel_scale=1.0)

        assert mask.centre == (-0.5, 0.0)

        mask = np.array([[True, True, True, True],
                         [True, False, False, True],
                         [False, True, True, True]])

        mask = msk.Mask(mask, pixel_scale=1.0)

        assert mask.centre == (-0.5, -0.5)


class TestMaskShapes:

    def test__mask_all_unmasked__5x5__input__all_are_false(self):

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(5, 5), pixel_scale=1.5, invert=False)

        assert mask.shape == (5, 5)
        assert (mask == np.array([[False, False, False, False, False],
                                 [False, False, False, False, False],
                                 [False, False, False, False, False],
                                 [False, False, False, False, False],
                                 [False, False, False, False, False]])).all()

        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)

    def test__mask_all_unmasked_inverted__5x5__input__all_are_true(self):

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(5, 5), pixel_scale=1, invert=True)

        assert mask.shape == (5, 5)
        assert (mask == np.array([[True, True, True, True, True],
                                 [True, True, True, True, True],
                                 [True, True, True, True, True],
                                 [True, True, True, True, True],
                                 [True, True, True, True, True]])).all()

        assert mask.origin == (0.0, 0.0)

    def test__mask_circular__compare_to_array_util(self):

        mask_util = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(5, 4), pixel_scale=2.7,
                                                                             radius_arcsec=3.5, centre=(0.0, 0.0))

        mask = msk.Mask.circular(shape=(5, 4), pixel_scale=2.7, radius_arcsec=3.5, centre=(0.0, 0.0))

        assert (mask == mask_util).all()
        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)

    def test__mask_circular__inverted__compare_to_array_util(self):

        mask_util = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(5, 4), pixel_scale=2.7,
                                                                             radius_arcsec=3.5, centre=(0.0, 0.0))

        mask = msk.Mask.circular(shape=(5, 4), pixel_scale=2.7, radius_arcsec=3.5, centre=(0.0, 0.0), invert=True)

        assert (mask == np.invert(mask_util)).all()
        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)

    def test__mask_annulus__compare_to_array_util(self):
        
        mask_util = util.mask_circular_annular_from_shape_pixel_scale_and_radii(shape=(5, 4), pixel_scale=2.7,
                                                                                    inner_radius_arcsec=0.8,
                                                                                    outer_radius_arcsec=3.5,
                                                                                    centre=(0.0, 0.0))

        mask = msk.Mask.circular_annular(shape=(5, 4), pixel_scale=2.7, inner_radius_arcsec=0.8, outer_radius_arcsec=3.5,
                                         centre=(0.0, 0.0))

        assert (mask == mask_util).all()
        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)

    def test__mask_annulus_inverted__compare_to_array_util(self):
        mask_util = util.mask_circular_annular_from_shape_pixel_scale_and_radii(shape=(5, 4), pixel_scale=2.7,
                                                                                inner_radius_arcsec=0.8,
                                                                                outer_radius_arcsec=3.5,
                                                                                centre=(0.0, 0.0))

        mask = msk.Mask.circular_annular(shape=(5, 4), pixel_scale=2.7, inner_radius_arcsec=0.8,
                                         outer_radius_arcsec=3.5,
                                         centre=(0.0, 0.0), invert=True)

        assert (mask == np.invert(mask_util)).all()
        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)

    def test__mask_anti_annulus__compare_to_array_util(self):
        mask_util = util.mask_circular_anti_annular_from_shape_pixel_scale_and_radii(shape=(9, 9), pixel_scale=1.2,
                                                                                         inner_radius_arcsec=0.8,
                                                                                         outer_radius_arcsec=2.2,
                                                                                         outer_radius_2_arcsec=3.0,
                                                                                         centre=(0.0, 0.0))

        mask = msk.Mask.circular_anti_annular(shape=(9, 9), pixel_scale=1.2, inner_radius_arcsec=0.8,
                                              outer_radius_arcsec=2.2, outer_radius_2_arcsec=3.0, centre=(0.0, 0.0))

        assert (mask == mask_util).all()
        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)

    def test__mask_anti_annulus_inverted__compare_to_array_util(self):
        mask_util = util.mask_circular_anti_annular_from_shape_pixel_scale_and_radii(shape=(9, 9), pixel_scale=1.2,
                                                                                         inner_radius_arcsec=0.8,
                                                                                         outer_radius_arcsec=2.2,
                                                                                         outer_radius_2_arcsec=3.0,
                                                                                         centre=(0.0, 0.0))

        mask = msk.Mask.circular_anti_annular(shape=(9, 9), pixel_scale=1.2, inner_radius_arcsec=0.8,
                                              outer_radius_arcsec=2.2, outer_radius_2_arcsec=3.0, centre=(0.0, 0.0),
                                              invert=True)

        assert (mask == np.invert(mask_util)).all()
        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)

    def test__mask_elliptical__compare_to_array_util(self):

        mask_util = util.mask_elliptical_from_shape_pixel_scale_and_radius(shape=(8, 5), pixel_scale=2.7,
                        major_axis_radius_arcsec=5.7, axis_ratio=0.4, phi=40.0, centre=(0.0, 0.0))

        mask = msk.Mask.elliptical(shape=(8, 5), pixel_scale=2.7,
                        major_axis_radius_arcsec=5.7, axis_ratio=0.4, phi=40.0, centre=(0.0, 0.0))

        assert (mask == mask_util).all()
        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)

    def test__mask_elliptical_inverted__compare_to_array_util(self):

        mask_util = util.mask_elliptical_from_shape_pixel_scale_and_radius(shape=(8, 5), pixel_scale=2.7,
                        major_axis_radius_arcsec=5.7, axis_ratio=0.4, phi=40.0, centre=(0.0, 0.0))

        mask = msk.Mask.elliptical(shape=(8, 5), pixel_scale=2.7,
                        major_axis_radius_arcsec=5.7, axis_ratio=0.4, phi=40.0, centre=(0.0, 0.0), invert=True)

        assert (mask == np.invert(mask_util)).all()
        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)

    def test__mask_elliptical_annular__compare_to_array_util(self):

        mask_util = util.mask_elliptical_annular_from_shape_pixel_scale_and_radius(shape=(8, 5), pixel_scale=2.7,
                        inner_major_axis_radius_arcsec=2.1, inner_axis_ratio=0.6, inner_phi=20.0,
                        outer_major_axis_radius_arcsec=5.7, outer_axis_ratio=0.4, outer_phi=40.0, centre=(0.0, 0.0))

        mask = msk.Mask.elliptical_annular(shape=(8, 5), pixel_scale=2.7,
                        inner_major_axis_radius_arcsec=2.1, inner_axis_ratio=0.6, inner_phi=20.0,
                        outer_major_axis_radius_arcsec=5.7, outer_axis_ratio=0.4, outer_phi=40.0, centre=(0.0, 0.0))

        assert (mask == mask_util).all()
        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)

    def test__mask_elliptical_annular_inverted__compare_to_array_util(self):

        mask_util = util.mask_elliptical_annular_from_shape_pixel_scale_and_radius(shape=(8, 5), pixel_scale=2.7,
                        inner_major_axis_radius_arcsec=2.1, inner_axis_ratio=0.6, inner_phi=20.0,
                        outer_major_axis_radius_arcsec=5.7, outer_axis_ratio=0.4, outer_phi=40.0, centre=(0.0, 0.0))

        mask = msk.Mask.elliptical_annular(shape=(8, 5), pixel_scale=2.7,
                        inner_major_axis_radius_arcsec=2.1, inner_axis_ratio=0.6, inner_phi=20.0,
                        outer_major_axis_radius_arcsec=5.7, outer_axis_ratio=0.4, outer_phi=40.0, centre=(0.0, 0.0),
                                           invert=True)

        assert (mask == np.invert(mask_util)).all()
        assert mask.origin == (0.0, 0.0)
        assert mask.centre == (0.0, 0.0)


class TestMaskMappings:

    def test__grid_to_pixel__compare_to_array_util(self):
        mask = np.array([[True, True, True],
                        [True, False, False],
                        [True, True, False]])

        mask = msk.Mask(mask, pixel_scale=7.0)

        grid_to_pixel_util = util.masked_grid_1d_index_to_2d_pixel_index_from_mask(mask)

        assert mask.masked_grid_index_to_pixel == pytest.approx(grid_to_pixel_util, 1e-4)

    def test__sub_grid_to_pixel__compare_to_array_util(self):

        mask = np.array([[True, True, True],
                        [True, False, False],
                        [True, True, False]])

        mask = msk.Mask(mask, pixel_scale=7.0)

        sub_grid_to_pixel_util = util.masked_sub_grid_1d_index_to_2d_sub_pixel_index_from_mask(mask=mask,
                                                                                               sub_grid_size=2)

        assert mask.masked_sub_grid_index_to_sub_pixel(sub_grid_size=2) == pytest.approx(sub_grid_to_pixel_util, 1e-4)

    def test__map_2d_array_to_masked_1d_array__compare_to_array_util(self):
        array_2d = np.array([[1, 2, 3],
                             [4, 5, 6],
                             [7, 8, 9],
                             [10, 11, 12]])

        mask = np.array([[True, False, True],
                        [False, False, False],
                        [True, False, True],
                        [True, True, True]])

        array_1d_util = mapping_util.map_2d_array_to_masked_1d_array_from_array_2d_and_mask(mask, array_2d)

        mask = msk.Mask(mask, pixel_scale=3.0)

        array_1d = mask.map_2d_array_to_masked_1d_array(array_2d)

        assert (array_1d == array_1d_util).all()


class TestMaskRegions:

    def test__blurring_mask_for_psf_shape__compare_to_array_util(self):
        
        mask = np.array([[True, True, True, True, True, True, True, True],
                        [True, False, True, True, True, False, True, True],
                        [True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True],
                        [True, False, True, True, True, False, True, True],
                        [True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True]])

        blurring_mask_util = util.mask_blurring_from_mask_and_psf_shape(mask=mask, psf_shape=(3, 3))

        mask = msk.Mask(mask, pixel_scale=1.0)
        blurring_mask = mask.blurring_mask_for_psf_shape(psf_shape=(3, 3))

        assert (blurring_mask == blurring_mask_util).all()

    def test__edge_image_pixels__compare_to_array_util(self):
        mask = np.array([[True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True],
                        [True, True, True, False, True, True, True],
                        [True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True]])

        edge_pixels_util = util.edge_pixels_from_mask(mask)

        mask = msk.Mask(mask, pixel_scale=3.0)

        assert mask.edge_pixels == pytest.approx(edge_pixels_util, 1e-4)

    def test__border_image_pixels__compare_to_array_util(self):

        mask = np.array([[False, False, False, False, False, False, False, True],
                         [False,  True,  True,  True,  True,  True, False, True],
                         [False,  True, False, False, False,  True, False, True],
                         [False,  True, False,  True, False,  True, False, True],
                         [False,  True, False, False, False,  True, False, True],
                         [False,  True,  True,  True,  True,  True, False, True],
                         [False, False, False, False, False, False, False, True]])

        border_pixels_util = util.border_pixels_from_mask(mask)

        mask = msk.Mask(mask, pixel_scale=3.0)

        assert mask.border_pixels == pytest.approx(border_pixels_util, 1e-4)


class TestMaskedGrid1d:

    def test__simple_grids(self):

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(3,3), pixel_scale=1.0)

        assert (mask.masked_grid_1d == np.array([[1., -1.], [1., 0.], [1., 1.],
                                                [0., -1.], [0., 0.], [0., 1.],
                                                [-1., -1.], [-1., 0.], [-1., 1.]])).all()

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(3,3), pixel_scale=1.0)
        mask[1,1] = True

        assert (mask.masked_grid_1d == np.array([[1., -1.], [1., 0.], [1., 1.],
                                                [0., -1.],            [0., 1.],
                                                [-1., -1.], [-1., 0.], [-1., 1.]])).all()

        mask = msk.Mask(array=np.array([[False, True],
                                        [True, False],
                                        [True, False]]), pixel_scale=1.0, origin=(3.0, -2.0))

        assert (mask.masked_grid_1d == np.array([[4., -2.5],
                                                             [3., -1.5],
                                                             [2., -1.5]])).all()

    def test__compare_to_grid_util(self):

        mask = msk.Mask.circular(shape=(4,7), radius_arcsec=4.0, pixel_scale=2.0, centre=(1.0, 5.0))

        masked_grid_1d_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(mask=mask,
                                                                                                 pixel_scales=(2.0, 2.0))

        assert (mask.masked_grid_1d == masked_grid_1d_util).all()


class TestZoomCentreAndOffet:

    def test__odd_sized_false_mask__centre_is_0_0__pixels_from_centre_are_0_0(self):

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(3,3), pixel_scale=1.0)
        assert mask.zoom_centre == (1.0, 1.0)
        assert mask.zoom_offset_pixels == (0, 0)

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(5,5), pixel_scale=1.0)
        assert mask.zoom_centre == (2.0, 2.0)
        assert mask.zoom_offset_pixels == (0, 0)

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(3,5), pixel_scale=1.0)
        assert mask.zoom_centre == (1.0, 2.0)
        assert mask.zoom_offset_pixels == (0, 0)

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(5,3), pixel_scale=1.0)
        assert mask.zoom_centre == (2.0, 1.0)
        assert mask.zoom_offset_pixels == (0, 0)

    def test__even_sized_false_mask__centre_is_0_0__pixels_from_centre_are_0_0(self):

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(4,4), pixel_scale=1.0)
        assert mask.zoom_centre == (1.5, 1.5)
        assert mask.zoom_offset_pixels == (0, 0)

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(6,6), pixel_scale=1.0)
        assert mask.zoom_centre == (2.5, 2.5)
        assert mask.zoom_offset_pixels == (0, 0)

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(4,6), pixel_scale=1.0)
        assert mask.zoom_centre == (1.5, 2.5)
        assert mask.zoom_offset_pixels == (0, 0)

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(6,4), pixel_scale=1.0)
        assert mask.zoom_centre == (2.5, 1.5)
        assert mask.zoom_offset_pixels == (0, 0)

    def test__mask_is_single_false__extraction_centre_is_central_pixel(self):

        mask = msk.Mask(array=np.array([[False, True, True],
                                        [True, True, True],
                                        [True, True, True]]), pixel_scale=1.0)
        assert mask.zoom_centre == (0, 0)
        assert mask.zoom_offset_pixels == (-1, -1)

        mask = msk.Mask(array=np.array([[True, True, False],
                                        [True, True, True],
                                        [True, True, True]]), pixel_scale=1.0)
        assert mask.zoom_centre == (0, 2)
        assert mask.zoom_offset_pixels == (-1, 1)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, True, True],
                                        [False, True, True]]), pixel_scale=1.0)
        assert mask.zoom_centre == (2, 0)
        assert mask.zoom_offset_pixels == (1, -1)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, True, True],
                                        [True, True, False]]), pixel_scale=1.0)
        assert mask.zoom_centre == (2, 2)
        assert mask.zoom_offset_pixels == (1, 1)

        mask = msk.Mask(array=np.array([[True, False, True],
                                        [True, True, True],
                                        [True, True, True]]), pixel_scale=1.0)
        assert mask.zoom_centre == (0, 1)
        assert mask.zoom_offset_pixels == (-1, 0)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [False, True, True],
                                        [True, True, True]]), pixel_scale=1.0)
        assert mask.zoom_centre == (1, 0)
        assert mask.zoom_offset_pixels == (0, -1)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, True, False],
                                        [True, True, True]]), pixel_scale=1.0)
        assert mask.zoom_centre == (1, 2)
        assert mask.zoom_offset_pixels == (0, 1)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, True, True],
                                        [True, False, True]]), pixel_scale=1.0)
        assert mask.zoom_centre == (2, 1)
        assert mask.zoom_offset_pixels == (1, 0)

    def test__mask_is_x2_false__extraction_centre_is_central_pixel(self):

        mask = msk.Mask(array=np.array([[False, True, True],
                                        [True, True, True],
                                        [True, True, False]]), pixel_scale=1.0)
        assert mask.zoom_centre == (1, 1)
        assert mask.zoom_offset_pixels == (0, 0)

        mask = msk.Mask(array=np.array([[False, True, True],
                                        [True, True, True],
                                        [False, True, True]]), pixel_scale=1.0)
        assert mask.zoom_centre == (1, 0)
        assert mask.zoom_offset_pixels == (0, -1)

        mask = msk.Mask(array=np.array([[False, True, False],
                                        [True, True, True],
                                        [True, True, True]]), pixel_scale=1.0)
        assert mask.zoom_centre == (0, 1)
        assert mask.zoom_offset_pixels == (-1, 0)

        mask = msk.Mask(array=np.array([[False, False, True],
                                        [True, True, True],
                                        [True, True, True]]), pixel_scale=1.0)
        assert mask.zoom_centre == (0, 0.5)
        assert mask.zoom_offset_pixels == (-1, -0.5)

    def test__rectangular_mask(self):

        mask = msk.Mask(array=np.array([[False, True, True, True],
                                        [True, True, True, True],
                                        [True, True, True, True]]), pixel_scale=1.0)

        assert mask.zoom_centre == (0, 0)
        assert mask.zoom_offset_pixels == (-1.0, -1.5)

        mask = msk.Mask(array=np.array([[True, True, True, True],
                                        [True, True, True, True],
                                        [True, True, True, False]]), pixel_scale=1.0)

        assert mask.zoom_centre == (2, 3)
        assert mask.zoom_offset_pixels == (1.0, 1.5)

        mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                        [True, True, True, True, True],
                                        [True, True, True, True, False]]), pixel_scale=1.0)

        assert mask.zoom_centre == (2, 4)
        assert mask.zoom_offset_pixels == (1, 2)

        mask = msk.Mask(array=np.array([[True, True, True, True, True, True, True],
                                        [True, True, True, True, True, True, True],
                                        [True, True, True, True, True, True, False]]), pixel_scale=1.0)

        assert mask.zoom_centre == (2, 6)
        assert mask.zoom_offset_pixels == (1, 3)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, True, True],
                                        [True, True, True],
                                        [True, True, True],
                                        [True, True, False]]), pixel_scale=1.0)

        assert mask.zoom_centre == (4, 2)
        assert mask.zoom_offset_pixels == (2, 1)

        mask = msk.Mask(array=np.array([[True, True, True],
                                        [True, True, True],
                                        [True, True, True],
                                        [True, True, True],
                                        [True, True, True],
                                        [True, True, True],
                                        [True, True, False]]), pixel_scale=1.0)

        assert mask.zoom_centre == (6, 2)
        assert mask.zoom_offset_pixels == (3, 1)


class TestMaskExtractor:

    def test__mask_extract_region__uses_the_limits_of_the_mask(self):

        mask = msk.Mask(array=np.array([[True,  True,  True, True],
                                        [True, False, False, True],
                                        [True, False, False, True],
                                        [True,  True,  True, True]]), pixel_scale=1.0)

        assert mask.zoom_region == [1, 3, 1, 3]

        mask = msk.Mask(array=np.array([[True,  True,  True, True],
                                        [True, False, False, True],
                                        [True, False, False, False],
                                        [True,  True,  True, True]]), pixel_scale=1.0)

        assert mask.zoom_region == [1, 3, 1, 4]

        mask = msk.Mask(array=np.array([[True,  True,  True, True],
                                        [True, False, False, True],
                                        [True, False, False, True],
                                        [True,  True, False, True]]), pixel_scale=1.0)

        assert mask.zoom_region == [1, 4, 1, 3]


        mask = msk.Mask(array=np.array([[True,  True,   True, True],
                                        [True, False,  False, True],
                                        [False, False, False, True],
                                        [True,  True,  True, True]]), pixel_scale=1.0)

        assert mask.zoom_region == [1, 3, 0, 3]

        mask = msk.Mask(array=np.array([[True, False,  True, True],
                                        [True, False, False, True],
                                        [True, False, False, True],
                                        [True,  True,  True, True]]), pixel_scale=1.0)

        assert mask.zoom_region == [0, 3, 1, 3]


class TestParse:

    def test__load_mask_from_fits__loads_mask(self):

        mask = msk.load_mask_from_fits(mask_path=test_data_dir + '3x3_ones.fits', pixel_scale=0.1)

        assert (mask == np.ones((3,3))).all()
        assert mask.pixel_scale == 0.1

    def test__output_mask_to_fits__outputs_mask(self):

        mask = msk.load_mask_from_fits(mask_path=test_data_dir + '3x3_ones.fits', pixel_scale=0.1)

        output_data_dir = "{}/../../test_files/array/output_test/".format(os.path.dirname(os.path.realpath(__file__)))
        if os.path.exists(output_data_dir):
            shutil.rmtree(output_data_dir)

        os.makedirs(output_data_dir)

        msk.output_mask_to_fits(mask=mask, mask_path=output_data_dir + 'mask.fits')

        mask = msk.load_mask_from_fits(mask_path=output_data_dir + 'mask.fits', pixel_scale=0.1)

        assert (mask == np.ones((3,3))).all()
        assert mask.pixel_scale == 0.1


class TestBinnedMaskFromMask:

    def test__compare_to_mask_util(self):

        mask = np.full(shape=(14, 19), fill_value=True)
        mask[1, 5] = False
        mask[6, 5] = False
        mask[4, 9] = False
        mask[11, 10] = False

        binned_up_mask_util = util.bin_up_mask_2d(mask_2d=mask, bin_up_factor=2)

        mask = msk.Mask(array=mask, pixel_scale=1.0)
        mask = mask.binned_up_mask_from_mask(bin_up_factor=2)
        assert (mask == binned_up_mask_util).all()
        assert mask.pixel_scale == 2.0

        binned_up_mask_util = util.bin_up_mask_2d(mask_2d=mask, bin_up_factor=3)

        mask = msk.Mask(array=mask, pixel_scale=2.0)
        mask = mask.binned_up_mask_from_mask(bin_up_factor=3)
        assert (mask == binned_up_mask_util).all()
        assert mask.pixel_scale == 6.0


class TestRescaledMaskFromMask(object):

    def test__mask_5x5_central_pixel__rescale_factor_is_1__returns_same_mask(self):

        mask = np.array([[True, True, True, True, True],
                         [True, True, True, True, True],
                         [True, True, False, True, True],
                         [True, True, True, True, True],
                         [True, True, True, True, True]])

        rescaled_mask = util.rescaled_mask_from_mask_and_rescale_factor(mask=mask, rescale_factor=1.0)

        assert (rescaled_mask == np.array([[True, True, True, True, True],
                                           [True, True, True, True, True],
                                           [True, True, False, True, True],
                                           [True, True, True, True, True],
                                           [True, True, True, True, True]])).all()

    def test__mask_5x5_central_pixel__rescale_factor_is_2__returns_10x10_mask_4_central_values(self):

        mask = np.array([[True, True, True, True, True],
                         [True, True, True, True, True],
                         [True, True, False, True, True],
                         [True, True, True, True, True],
                         [True, True, True, True, True]])

        rescaled_mask = util.rescaled_mask_from_mask_and_rescale_factor(mask=mask, rescale_factor=2.0)

        assert (rescaled_mask == np.array([[True, True, True, True, True, True, True, True, True, True],
                                           [True, True, True, True, True, True, True,  True, True, True],
                                           [True, True, True, True, True, True, True, True, True, True],
                                           [True, True, True, False, False, False, False, True, True, True],
                                           [True, True, True, False, False, False, False, True, True, True],
                                           [True, True, True, False, False, False, False, True, True, True],
                                           [True, True, True, False, False, False, False, True, True, True],
                                           [True, True, True, True, True, True, True, True, True, True],
                                           [True, True, True, True, True, True, True, True, True, True],
                                           [True, True, True, True, True, True, True, True, True, True]])).all()

    def test__same_as_above__off_centre_pixels(self):

        mask = np.array([[True, True, True, True, True],
                         [True, False, True, True, True],
                         [True, True, True, True, True],
                         [True, True, True, False, True],
                         [True, True, True, True, True]])

        rescaled_mask = util.rescaled_mask_from_mask_and_rescale_factor(mask=mask, rescale_factor=2.0)

        assert (rescaled_mask == np.array([[True, True, True, True, True, True, True, True, True, True],
                                           [True, False, False, False, False, True, True,  True, True, True],
                                           [True, False, False, False, False, True, True, True, True, True],
                                           [True, False, False, False, False, True, True, True, True, True],
                                           [True, False, False, False, False, True, True, True, True, True],
                                           [True, True, True, True, True, False, False, False, False, True],
                                           [True, True, True, True, True, False, False, False, False, True],
                                           [True, True, True, True, True, False, False, False, False, True],
                                           [True, True, True, True, True, False, False, False, False, True],
                                           [True, True, True, True, True, True, True, True, True, True]])).all()

    def test__mask_4x3_two_central_pixels__rescale_near_1__returns_slightly_different_masks(self):

        mask = np.array([[True, True, True, True, True],
                         [True, True, True, True, True],
                         [True, True, False, True, True],
                         [True, True, False, True, True],
                         [True, True, True, True, True],
                         [True, True, True, True, True]])

        rescaled_mask = util.rescaled_mask_from_mask_and_rescale_factor(mask=mask, rescale_factor=1.2)

        assert (rescaled_mask == np.array([[True, True, True, True, True, True],
                                           [True, True, True, True, True, True],
                                           [True, True, False, False, True, True],
                                           [True, True, False, False, True, True],
                                           [True, True, False, False, True, True],
                                           [True, True, True, True, True, True],
                                           [True, True, True, True, True, True]])).all()

        rescaled_mask = util.rescaled_mask_from_mask_and_rescale_factor(mask=mask, rescale_factor=0.8)

        assert (rescaled_mask == np.array([[True, True, True, True],
                                           [True, False, False, True],
                                           [True, False, False, True],
                                           [True, False, False, True],
                                           [True, True, True, True]])).all()

    def test__mask_3x4_two_central_pixels__rescale_near_1__returns_slightly_different_masks(self):

        mask = np.array([[True, True, True, True, True, True],
                         [True, True, True, True, True, True],
                         [True, True, False, False, True, True],
                         [True, True, True, True, True, True],
                         [True, True, True, True, True, True]])

        rescaled_mask = util.rescaled_mask_from_mask_and_rescale_factor(mask=mask, rescale_factor=1.2)

        assert (rescaled_mask == np.array([[True, True, True, True, True, True, True],
                                           [True, True, True, True, True, True, True],
                                           [True, True, False, False, False, True, True],
                                           [True, True, False, False, False, True, True],
                                           [True, True, True, True, True, True, True],
                                           [True, True, True, True, True, True, True]])).all()

        rescaled_mask = util.rescaled_mask_from_mask_and_rescale_factor(mask=mask, rescale_factor=0.8)

        assert (rescaled_mask == np.array([[True, True, True, True, True],
                                           [True, False, False, False, True],
                                           [True, False, False, False, True],
                                           [True, True, True, True, True]])).all()


class TestEdgeBuffedMaskFromMask(object):

    def test__5x5_mask_false_centre_pixel__3x3_falses_in_centre_of_edge_buffed_mask(self):

        mask = np.array([[True, True, True, True, True],
                         [True, True, True, True, True],
                         [True, True, False, True, True],
                         [True, True, True, True, True],
                         [True, True, True, True, True]])

        edge_buffed_mask = util.edge_buffed_mask_from_mask(mask=mask)

        assert (edge_buffed_mask == np.array([[True, True, True, True, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True, False, False, False, True],
                                              [True, True, True, True, True]])).all()

    def test__5x5_mask_false_offset_pixel__3x3_falses_in_centre_of_edge_buffed_mask(self):

        mask = np.array([[True, True, True, True, True],
                         [True, True, True, True, True],
                         [True, True, True, True, True],
                         [True, True, True, False, True],
                         [True, True, True, True, True]])

        edge_buffed_mask = util.edge_buffed_mask_from_mask(mask=mask)

        assert (edge_buffed_mask == np.array([[True, True, True, True, True],
                                              [True, True, True, True, True],
                                              [True, True, False, False, False],
                                              [True, True, False, False, False],
                                              [True, True, False, False, False]])).all()

    def test__mask_4x3__buffed_mask_same_shape(self):

        mask = np.array([[True, True, True, True, True],
                         [True, True, True, True, True],
                         [True, True, False, True, True],
                         [True, True, False, True, True],
                         [True, True, True, True, True],
                         [True, True, True, True, True]])

        edge_buffed_mask = util.edge_buffed_mask_from_mask(mask=mask)

        assert (edge_buffed_mask == np.array([[True, True, True, True, True],
                                           [True,  False, False, False, True],
                                           [True,  False, False, False, True],
                                           [True,  False, False, False, True],
                                           [True,  False, False, False, True],
                                           [True, True, True, True, True]])).all()

    def test__mask_3x4_two_central_pixels__rescale_near_1__returns_slightly_different_masks(self):

        mask = np.array([[True, True, True, True, True, True],
                         [True, True, True, True, True, True],
                         [True, True, False, False, True, True],
                         [True, True, True, True, True, True],
                         [True, True, True, True, True, True]])

        edge_buffed_mask = util.edge_buffed_mask_from_mask(mask=mask)

        assert (edge_buffed_mask == np.array([[True, True, True, True, True, True],
                                              [True, False, False, False, False, True],
                                              [True, False, False, False, False, True],
                                              [True, False, False, False, False, True],
                                              [True, True, True, True, True, True]])).all()


from autolens.data.array import grids
from autolens.model.profiles import mass_profiles as mp

# This numerics test determines the range of plausible values a power-law mass profile can take and not crash due to
# numerical integration. This informs us the limits config file for this profile.

# In this test, we'll pass a regular grid to a power-law profile and compute deflection angles. We will check if
# the pyquad integrator crashes for certain values of (y,x) coordinates.
regular_grid = grids.RegularGrid.from_shape_and_pixel_scale(shape=(25, 25), pixel_scale=0.005)

y_centres = [0.01, 0.005, 0.001, 0.0001, 0.00000001, 0.0]

for y in y_centres:

    print('Normal Isothermal (centre offset = {:.8f})'.format(y))
    power_law = mp.EllipticalIsothermal(centre=(y, 0.0), axis_ratio=0.8, phi=0.0, einstein_radius=1.0)
    power_law.deflections_from_grid(grid=regular_grid, grid_radial_minimum=1.0e-8)

for y in y_centres:

    print('Axis Ratio Max Isothermal (centre offset = {:.8f})'.format(y))
    power_law = mp.EllipticalIsothermal(centre=(y, 0.0), axis_ratio=0.99999999, phi=0.0, einstein_radius=1.0)
    power_law.deflections_from_grid(grid=regular_grid, grid_radial_minimum=1.0e-8)

for y in y_centres:

    print('Axis Ratio 0.01 Isothermal (centre offset = {:.8f})'.format(y))
    power_law = mp.EllipticalIsothermal(centre=(y, 0.0), axis_ratio=0.005, phi=0.0, einstein_radius=1.0)
    power_law.deflections_from_grid(grid=regular_grid, grid_radial_minimum=1.0e-8)

for y in y_centres:

    print('Slope Low Isothermal (centre offset = {:.8f})'.format(y))
    power_law = mp.EllipticalIsothermal(centre=(y, 0.0), axis_ratio=0.8, phi=0.0, einstein_radius=1.0)
    power_law.deflections_from_grid(grid=regular_grid, grid_radial_minimum=1.0e-8)

for y in y_centres:

    print('Slope High Normal Isothermal (centre offset = {:.8f})'.format(y))
    power_law = mp.EllipticalIsothermal(centre=(y, 0.0), axis_ratio=0.8, phi=0.0, einstein_radius=1.0)
    power_law.deflections_from_grid(grid=regular_grid, grid_radial_minimum=1.0e-8)


from autolens.model.profiles import mass_profiles as mp

def summarize_mass_profile(summary_file, mass_profile, critical_surface_mass_density,
                           cosmic_average_mass_density_arcsec, radii):

    summary_file.write('Mass Profile = ' + mass_profile.__class__.__name__ + '\n\n')
    summarize_einstein_radius_and_mass(summary_file=summary_file, mass_profile=mass_profile,
                                       critical_surface_mass_density=critical_surface_mass_density)
    summarize_mass_within_radii(summary_file=summary_file, mass_profile=mass_profile,
                                critical_surface_mass_density=critical_surface_mass_density,
                                radii=radii)

    if isinstance(mass_profile, mp.SphericalTruncatedNFWChallenge):
        summarize_truncated_nfw_challenge_mass_profile(summary_file=summary_file, truncated_nfw_challenge=mass_profile)



def summarize_einstein_radius_and_mass(summary_file, mass_profile, critical_surface_mass_density):
    """ Summarize the mass at the Einstein radius of the mass profile.

    Parameters
    -----------
    summary_file : file
        The summary file the mass-profiles information is written to.
    mass_profile : model.profiles.mass_profiles.MassProfile
        The mass profile instance whose mass within radii are summarized.
    critical_surface_mass_density : float
        The critical surface mass density of the lensing system, required to convert the mass to physical units.
    """

    einstein_mass = mass_profile.mass_within_circle_in_mass_units(
        radius=mass_profile.einstein_radius, critical_surface_mass_density=critical_surface_mass_density)

    summary_file.write('Einstein Radius = {:.2f}"\n'.format(mass_profile.einstein_radius))
    summary_file.write('Mass within Einstein Radius = {:.4e} solMass\n'.format(einstein_mass))

def summarize_mass_within_radii(summary_file, mass_profile, critical_surface_mass_density, radii):
    """ Summarize the mass within a set of input radii of a given mass profile.

    Parameters
    -----------
    summary_file : file
        The summary file the mass-profiles information is written to.
    mass_profile : model.profiles.mass_profiles.MassProfile
        The mass profile instance whose mass within radii are summarized.
    radii : [float]
        The radii at which the mass of the mass profile is output to the summary file.
    critical_surface_mass_density : float
        The critical surface mass density of the lensing system, required to convert the mass to physical units.
    """

    for radius in radii:

        mass = mass_profile.mass_within_circle_in_mass_units(
            radius=radius, critical_surface_mass_density=critical_surface_mass_density)

        summary_file.write('Mass within {:.2f}" = {:.4e} solMass\n'.format(radius, mass))

def summarize_nfw_mass_profile(summary_file, nfw, critical_surface_mass_density_arcsec,
                               cosmic_average_mass_density_arcsec):

    rho_at_scale_radius = \
        nfw.rho_at_scale_radius(critical_surface_mass_density_arcsec=critical_surface_mass_density_arcsec)

    delta_concentration = \
        nfw.delta_concentration(critical_surface_mass_density_arcsec=critical_surface_mass_density_arcsec,
                                cosmic_average_mass_density_arcsec=cosmic_average_mass_density_arcsec)

    concentration = nfw.concentration(critical_surface_mass_density_arcsec=critical_surface_mass_density_arcsec,
                                      cosmic_average_mass_density_arcsec=cosmic_average_mass_density_arcsec)

    radius_at_200 = nfw.radius_at_200(critical_surface_mass_density_arcsec=critical_surface_mass_density_arcsec,
                                      cosmic_average_mass_density_arcsec=cosmic_average_mass_density_arcsec)

    mass_at_200 = nfw.mass_at_200(critical_surface_mass_density_arcsec=critical_surface_mass_density_arcsec,
                                  cosmic_average_mass_density_arcsec=cosmic_average_mass_density_arcsec)

    summary_file.write('Rho at scale radius = {:.2f}\n'.format(rho_at_scale_radius))
    summary_file.write('Delta concentration = {:.2f}\n'.format(delta_concentration))
    summary_file.write('Concentration = {:.2f}\n'.format(concentration))
    summary_file.write('Radius at 200x cosmic average density = {:.2f}"\n'.format(radius_at_200))
    summary_file.write('Mass at 200x cosmic average density = {:.2f} solMass\n'.format(mass_at_200))

def summarize_truncated_nfw_mass_profile(summary_file, truncated_nfw, critical_surface_mass_density_arcsec,
                                         cosmic_average_mass_density_arcsec):

    summarize_nfw_mass_profile(summary_file=summary_file, nfw=truncated_nfw,
                               critical_surface_mass_density_arcsec=critical_surface_mass_density_arcsec,
                               cosmic_average_mass_density_arcsec=cosmic_average_mass_density_arcsec)

    mass_at_truncation_radius = truncated_nfw.mass_at_truncation_radius(
        critical_surface_mass_density_arcsec=critical_surface_mass_density_arcsec,
        cosmic_average_mass_density_arcsec=cosmic_average_mass_density_arcsec)

    summary_file.write('Mass at truncation radius = {:.2f} solMass\n'.format(mass_at_truncation_radius))

def summarize_truncated_nfw_challenge_mass_profile(summary_file, truncated_nfw_challenge):

    summarize_truncated_nfw_mass_profile(summary_file=summary_file, truncated_nfw=truncated_nfw_challenge,
                                         critical_surface_mass_density_arcsec=1940654909.4133248,
                                         cosmic_average_mass_density_arcsec=262.30319684750657)


from autofit.tools import path_util
from autofit.optimize import non_linear as nl
from autofit.mapper import prior
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp

import os

# In this pipeline, we'll perform an initializer analysis which fits an image with a source galaxy and no lens light
# component. The pipeline is as follows:

# Phase 1:

# Description: Initializes the lens mass model and source light profile.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: None

def make_pipeline(phase_folders=None, phase_tagging=True, sub_grid_size=2, bin_up_factor=None, positions_threshold=None,
                  inner_mask_radii=None, interp_pixel_scale=None):

    pipeline_name = 'pipeline_init__lens_sie_shear_source_sersic'

    phase_folders = path_util.phase_folders_from_phase_folders_and_pipeline_name(phase_folders=phase_folders,
                                                                                pipeline_name=pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy (y,x) centre such that we assume the image is centred around the lens galaxy.

    class LensSourcePhase(ph.LensSourcePlanePhase):

        def pass_priors(self, results):

            ## Lens Mass, move centre priors to centre of image ###

            self.lens_galaxies.lens.mass.centre_0 = prior.GaussianPrior(mean=0.0, sigma=0.3)
            self.lens_galaxies.lens.mass.centre_1 = prior.GaussianPrior(mean=0.0, sigma=0.3)

    phase1 = LensSourcePhase(phase_name='phase_1_lens_sie_shear_source_sersic', phase_folders=phase_folders,
                             phase_tagging=phase_tagging,
                             lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal,
                                                                    shear=mp.ExternalShear)),
                             source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalSersic)),
                             sub_grid_size=sub_grid_size, bin_up_factor=bin_up_factor,
                             positions_threshold=positions_threshold, inner_mask_radii=inner_mask_radii,
                             interp_pixel_scale=interp_pixel_scale,
                             optimizer_class=nl.MultiNest)

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    return pipeline.PipelineImaging(pipeline_name, phase1)